/* FLUTPPC.CPP - File containing the implementation of functions in
   FLUTPPC.H
   Eduardo T Ochiai revised 1999/03/03 ver 1.1

   Note that , after every arithmetic operation, the number is normalized
   to the form :
        d.dd...dE[+/-]dd..d
   i.e. with only one integer digit      
 */

#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iostream.h>
#include "flutppc.h"

Ereal zero(ZERO),
      um(UM),
      eps(EPS),
      pi1_2(PI1_2),
      pi(PI),
      pi3_2(PI3_2),
      pi_2(PI_2),
      e(E);

// Sets number to zero
inline void Ereal::zera()
{
e=0;
memset(d,0,max);
}

// Shifts number n digits to the left, without any change to the exponent (e)
inline void Ereal::desl_esq(int n)
{
if (n<max_tot) {
memmove(&(d[1]),&(d[n+1]),max_tot-n);
memset(&(d[max_tot-n+1]),0,n);
}
else zera();
}

// Shifts number n digits to the right, without any change to the exponent (e)
inline void Ereal::desl_dir(int n)
{
if (n<=max_tot-1) {
memmove(&(d[n+1]),&(d[1]),max_tot-n);
memset(&(d[1]),0,n);
}
else zera();
}

// Default constructor
inline Ereal::Ereal()
{
 d=new char[max];

 if (!d) {
   throw (TMathError("Out of memory"));
 }

 e=0;
 memset(d,0,max);
}

// Copy constructor
inline Ereal::Ereal(const Ereal &x)
{
 d=new char[max];

 if (!d) {
   throw (TMathError("Out of memory"));
 }

 e=x.e;
 memcpy(d,x.d,max);
}

// Constructor passing a string as parameter 
Ereal::Ereal(char *number)
{
char a,h;
char *r;
int i,j,p;

d=new char[max];

if (!d) {
  throw (TMathError("Out of memory"));
}

e=0;
memset(d,0,max);

i=0; p=0;
d[0]=0;
if ((number[i]=='+') || (number[i]=='-'))
 {
 if (number[i]=='+') d[0]=1;
 else d[0]=-1;
 i++;
 }
j=1; a='\0';
do {
 a=number[i];
 if ((a!='0') && (d[0]==0)) d[0]=1;
 d[j]=a-'0';
 j++; i++;
 } while ((number[i]!='E') && (number[i]!='.') && (number[i]!='\0'));
if (d[1]!=0) p=j-2;
else p=-1;
h=FALSE;
if (number[i]=='.')
 {
 i++;
 do {
 a=number[i];
 if ((a!='0') && (d[0]==0))
  {
  d[0]=1;
  h=TRUE;
  }
 if ((a=='0') && (d[1]==0) && (!h)) p--;
 d[j]=a-'0';
 j++; i++;
 } while ((number[i]!='E') && (number[i]!='\0'));
 }
if (number[i]=='E')
 {
 r=strstr(number,"E");
 r++;
 e=atoi(r);
 }
if (d[1]!=0) desl_dir(2);
else if (d[1]==0)
 {
 if (-p-2>0) desl_esq(-p-2);
 else desl_dir(p+2);
 }
e=e+p;
}

// Destructor (note that (oops) the 'delete' was omitted accidentally
// in the last version (1.0))
Ereal::~Ereal()
{
delete [] d;
}

inline Ereal& Ereal::operator=(const Ereal& x)
{
 e=x.e;
 memmove(d,x.d,max);
 return *this;
}

Ereal r(char *number)
{
char a,h;
char *s;
int i,j,p;
Ereal x;

x.zera();
i=0; p=0;
x.d[0]=0;
if ((number[i]=='+') || (number[i]=='-'))
 {
 if (number[i]=='+') x.d[0]=1;
 else x.d[0]=-1;
 i++;
 }
j=1; a='\0';
do {
 a=number[i];
 if ((a!='0') && (x.d[0]==0)) x.d[0]=1;
 x.d[j]=a-'0';
 j++; i++;
 } while ((number[i]!='E') && (number[i]!='.') && (number[i]!='\0'));
if (x.d[1]!=0) p=j-2;
else p=-1;
h=FALSE;
if (number[i]=='.')
 {
 i++;
 do {
 a=number[i];
 if ((a!='0') && (x.d[0]==0))
  {
  x.d[0]=1;
  h=TRUE;
  }
 if ((a=='0') && (x.d[1]==0) && (!h)) p--;
 x.d[j]=a-'0';
 j++; i++;
 } while ((number[i]!='E') && (number[i]!='\0'));
 }
if (number[i]=='E')
 {
 s=strstr(number,"E");
 s++;
 x.e=atoi(s);
 }
if (x.d[1]!=0) x.desl_dir(2);
else if (x.d[1]==0)
 {
 if (-p-2>0) x.desl_esq(-p-2);
 else x.desl_dir(p+2);
 }
x.e=x.e+p;
return x;
}

int Ereal::tam_int()
{
int i;
i=1;
while ((d[i]==0) && (i<=max_int)) i++;
return (max_int-i+1);
}

int Ereal::tam_dec()
{
int i;
i=max_tot;
while ((!d[i]) && (i>max_int)) i--;
return(i-max_int);
}

int operator==(Ereal& x, Ereal& y)
{
if (x.e!=y.e) return FALSE;
if (memicmp(x.d,y.d,max)!=0) return FALSE;
return TRUE;
}

int operator!=(Ereal& x, Ereal& y)
{
if (!(x==y)) return TRUE;
else return FALSE;
}

int operator>(Ereal& x, Ereal& y)
{
int tx,ty;
int f,ig;

ig=(x==y);
if ((x.d[0]*y.d[0]<=0) && ((x.d[0]!=0) || (y.d[0]!=0)))
 {
 if (x.d[0]<y.d[0]) f=FALSE;
	       else f=TRUE;
 }
else
 {
 if (x.e<y.e) f=FALSE;
 else if (x.e>y.e) f=TRUE;
 else
  {
  f=TRUE;
  tx=x.tam_int();
  ty=y.tam_int();
  if (tx>ty) f=TRUE;
  else if (tx<ty) f=FALSE;
  else if ((tx==ty) && (!ig)) {
   if (memicmp(&(x.d[1]),&(y.d[1]),max-1)<0) f=FALSE;
  }
  else f=FALSE;
  }
 if ((x.d[0]<0) && (!ig)) f=!f;
 }
 return f;
}

int operator<(Ereal& x, Ereal& y)
{
if ((x>y) || (x==y)) return FALSE;
		     else return TRUE;
}

int operator>=(Ereal& x, Ereal& y)
{
if ((x>y) || (x==y)) return TRUE;
		else return FALSE;
}

int operator<=(Ereal& x, Ereal& y)
{
if ((x<y) || (x==y)) return TRUE;
		else return FALSE;
}

Ereal Ereal::operator+()
{
 return *this;
}

Ereal Ereal::operator-()
{
 this->d[0] = -this->d[0];
 return *this;
}

Ereal& Ereal::operator+=(Ereal& x)
{
 *this=*this+x;
 return *this;
}

Ereal& Ereal::operator-=(Ereal& x)
{
 *this=*this-x;
 return *this;
}

Ereal& Ereal::operator*=(Ereal& x)
{
 *this=(*this)*x;
 return *this;
}

Ereal& Ereal::operator/=(Ereal& x)
{
 *this=(*this)/x;
 return *this;
}

// Swaps x and y
void troca(Ereal& x, Ereal& y)
{
Ereal z;
z=y;
y=x;
x=z;
}

// Formats x directing output to "saida"
char *texto(Ereal& x, char *saida)
{
int i,j,k;
char *tmp=saida;
j=x.tam_dec();
k=x.tam_int();
if (x.d[0]==0) sprintf(saida++,"0\0");
else {
if (x.d[0]<0) sprintf(saida++,"-");
else if (x.d[0]>0) sprintf(saida++,"+");
for (i=max_int-k+1;i<=max_int;++i) sprintf(saida++,"%u",x.d[i]);
if (j!=0) sprintf(saida++,".");
for (i=max_int+1;i<=max_int+j;++i) sprintf(saida++,"%u",x.d[i]);
if (x.d[0]==0) sprintf(saida++,"0");
sprintf(saida++,"E");
sprintf(saida++,"%+i\0",x.e);
}
return(tmp);
}

// Reads number from console using gets()
void le_numero(Ereal *x)
{
char *k;
char a,h;
int i,j,p;

if ((k=new char[max+9])==NULL) exit(1);
(*x).zera();
k[0]='\0';
gets(k);
i=0; p=0;
x->d[0]=0;
if ((k[i]=='+') || (k[i]=='-'))
 {
 if (k[i]=='+') x->d[0]=1;
 else x->d[0]=-1;
 i++;
 }
j=1; a='\0';
do {
 a=k[i];
 if ((a!='0') && (x->d[0]==0)) x->d[0]=1;
 x->d[j]=a-'0';
 j++; i++;
 } while ((k[i]!='E') && (k[i]!='.') && (k[i]!='\0'));
if (x->d[1]!=0) p=j-2;
else p=-1;
h=FALSE;
if (k[i]=='.')
 {
 i++;
 do {
 a=k[i];
 if ((a!='0') && (x->d[0]==0))
  {
  x->d[0]=1;
  h=TRUE;
  }
 if ((a=='0') && (x->d[1]==0) && (!h)) p--;
 x->d[j]=a-'0';
 j++; i++;
 } while ((k[i]!='E') && (k[i]!='\0'));
 }
if (k[i]=='E')
 {
 i++;
 x->e=atoi(&(k[i]));
 }
if (x->d[1]!=0) (*x).desl_dir(2);
else if (x->d[1]==0)
 {
 if (-p-2>0) (*x).desl_esq(-p-2);
 else (*x).desl_dir(p+2);
 }
x->e=x->e+p;
delete [] k;
}

// Shifts number 1 digit left and fills last position with digit 'k'
inline void Ereal::desce_alg(char k)
{
if ((d[0]==0) && (k!=0)) d[0]=1;
desl_esq(1);
d[max_tot]=k;
}

// Shifts number 2 digits left and fills last 2 position2 with digits 'k' and 'j'
inline void Ereal::desce_2alg(char k ,char j)
{
if ((d[0]==0) && ((k!=0) || (j!=0))) d[0]=1;
desl_esq(2);
d[max_tot-1]=k;
d[max_tot]=j;
}

Ereal operator-(Ereal& x, Ereal& y);

Ereal operator+(Ereal& x, Ereal& y)
{
Ereal u,v,z;
int i;
char r,s;
int k,t;
char sinalx,sinaly;

z.zera();
u.zera();
v.zera();
u=x;
v=y;

if (u.d[0]*v.d[0]<0)
  {
  sinalx=u.d[0];
  sinaly=v.d[0];
  u.d[0]=1;
  v.d[0]=1;
  if (u>v)
   {
   z=u-v;
   z.d[0]=sinalx;
   }
  else if (u<v)
   {
   z=v-u;
   z.d[0]=sinaly;
   }
  }
else if (u.d[0]*v.d[0]>0)
  {
  if (u.e>v.e) {
	     k=u.e-v.e;
	     z.e=u.e;
	     v.desl_dir(k);
	     }
  else if (u.e<v.e)
     {
     k=v.e-u.e;
     z.e=v.e;
     u.desl_dir(k);
     }
  else z.e=u.e;
  if ((u.d[0]==0) && (v.d[0]==0)) z.d[0]=0;
  else if (u.d[0]==0) z.d[0]=v.d[0];
  else if (v.d[0]==0) z.d[0]=u.d[0];
  else z.d[0]=u.d[0];
  r=0;
  k=u.tam_dec();
  t=v.tam_dec();
  if (k>t) i=max_int+k ;
  else i=max_int+t;
  k=u.tam_int();
  t=v.tam_int();
  if (t>k) k=t;
  while (i>=max_int-k)
   {
   s=u.d[i]+v.d[i]+r;
   z.d[i]=s % 10;
   r= s / 10;
   i--;
   }
  }
else {
 if ((v.d[0]==0) && (u.d[0]==0)) z.zera();
 else if (u.d[0]==0) z=v;
 else if (v.d[0]==0) z=u;
 }

if (z.d[0]!=0)
{
if (z.tam_int()>1) do
		      {
		      z.desl_dir(1);
		      z.e=z.e+1;
		      } while (z.tam_int()>1);
else if (z.tam_int()<1) do
		      {
		      z.desl_esq(1);
		      z.e=z.e-1;
		      } while (z.tam_int()<1);
}
return(z);
}

Ereal operator-(Ereal& x, Ereal& y)
{
int i;
char r,s;
int k,t;
int sinal;
Ereal u,v,z;

u.zera();
v.zera();
u=x;
v=y;
z.zera();

if (u.d[0]*v.d[0]<=0)
  {
  if ((u.d[0]==0) && (v.d[0]==0)) z.d[0]=0;
  else if (u.d[0]==0)
     {
     z=v;
     z.d[0]=-v.d[0];
     }
  else if (v.d[0]==0) z=u;
  else
     {
     sinal=u.d[0];
     u.d[0]=1;
     v.d[0]=1;
     z=u+v;
     if (sinal>0) z.d[0]=1;
     else z.d[0]=-1;
     }
  }
else
  {
  sinal=u.d[0];
  if (u.e>v.e) {
	     k=u.e-v.e;
	     z.e=u.e;
	     v.desl_dir(k);
	     }
  else if (u.e<v.e)
     {
     k=v.e-u.e;
     z.e=v.e;
     u.desl_dir(k);
     }
  else z.e=u.e;
//  sinal=u.d[0];
  u.d[0]=abs(u.d[0]); //  u.d[0]=1;
  v.d[0]=abs(v.d[0]); //  v.d[0]=1;
  if (u>v) z.d[0]=sinal;
  else if (u<v)
    {
    z.d[0]=-sinal;
    troca(u,v);
    }
  else if (u==v) z.d[0]=0;
  r=0;
  k=u.tam_dec();
  t=v.tam_dec();
  if (k>t) i=max_int+k;
  else i=max_int+t;
  k=u.tam_int();
  t=v.tam_int();
  if (t>k) k=t;
  while (i>=max_int-k)
   {
   s=u.d[i]-v.d[i]-r;
   z.d[i]=(10+s) % 10;
   r= ! ((10+s) / 10);
   i--;
   }
  }
if (z.d[0]!=0)
{
if (z.tam_int()>1) do
		      {
		      z.desl_dir(1);
		      z.e=z.e+1;
		      } while (z.tam_int()>1);
else if (z.tam_int()<1) do
		      {
		      z.desl_esq(1);
		      z.e=z.e-1;
		      } while (z.tam_int()<1);
}
return(z);
}

// This is an auxiliary function used mainly in division
Ereal subt_f(Ereal x, Ereal y)
{
int i;
char r,s;
int k,t;
int sinal;
Ereal z;

z.zera();
if (x.e>y.e) {
	       k=x.e-y.e;
	       z.e=x.e;
	       y.desl_dir(k);
	       }
else if (x.e<y.e)
     {
     k=y.e-x.e;
     z.e=y.e;
     x.desl_dir(k);
     }

if (x.d[0]*y.d[0]<=0)
  {
  if ((x.d[0]==0) && (x.d[0]==0)) z.d[0]=0;
  else if (x.d[0]==0)
     {
     z=y;
     z.d[0]=-y.d[0];
     }
  else if (y.d[0]==0) z=x;
  else
     {
     sinal=x.d[0];
     x.d[0]=1;
     y.d[0]=1;
     z=x+y;
     if (sinal>0) z.d[0]=1;
     else z.d[0]=-1;
     }
  }
else
  {
  sinal=x.d[0];
  x.d[0]=1;
  y.d[0]=1;
  if (x>y) z.d[0]=sinal;
  else if (x<y)
    {
    z.d[0]=-sinal;
    troca(x,y);
    }
  else if (x==y) z.d[0]=0;
  r=0;
  k=x.tam_dec();
  t=y.tam_dec();
  if (k>t) i=max_int+k;
  else i=max_int+t;
  k=x.tam_int();
  t=y.tam_int();
  if (t>k) k=t;
  while (i>=max_int-k)
   {
   s=x.d[i]-y.d[i]-r;
   z.d[i]=(10+s) % 10;
   r= ! ((10+s) / 10);
   i--;
   }
  }
return(z);
}

Ereal operator*(Ereal& x, Ereal& y)
{
int i,j,k;
int ix,dx,iy;
char r,r1,s,s1;
char *temp;
char *temp1;
Ereal u,v,z;

if ((temp=new char[2*max_tot+1])==NULL)
   throw (TMathError("Out of memory"));
if ((temp1=new char[2*max_tot+1])==NULL) {
   delete[] temp;    throw (TMathError("Out of memory"));
}

u.zera();
v.zera();
z.zera();
if ((x.d[0]==0) || (y.d[0]==0)) {
  delete[] temp;
  delete[] temp1;
  return(z);
}
u=x;
v=y;
for (i=0;i<=2*max_tot;++i)
	 {
	 temp[i]=0;
	 temp1[i]=0;
	 }
z.d[0]=u.d[0]*v.d[0];
if((u.e/2+v.e/2)>16382 || (u.e/2+v.e/2)<-16382) {
  delete[] temp;
  delete[] temp1;
  throw (TMathError("Number out of range."));
}
z.e=u.e+v.e;
i=max_int+v.tam_dec();
iy=v.tam_int();
dx=u.tam_dec();
ix=u.tam_int();
while (i>=max_int-iy+1)
 {
 r=0; s=0; r1=0;
 j=max_int+dx;
 while (j>=max_int-ix)
  {
  s=v.d[i]*u.d[j]+r;
  temp[i+j]=s % 10;
  r= s / 10;
  s1=temp[i+j]+temp1[i+j]+r1;
  temp[i+j]=s1 % 10;
  r1= s1 / 10;
  j--;
  }
 temp[i+j]=temp[i+j]+temp1[i+j]+r1;
 for (k=1;k<=max_int+dx+i;++k)
	 {
	 temp1[k]=temp[k];
	 temp[k]=0;
	 }
 i--;
}
memcpy(&(z.d[1]),&(temp1[max_int+1]),max_tot);
if (z.d[0]!=0)
{
if (z.tam_int()>1) do
		      {
		      z.desl_dir(1);
		      z.e=z.e+1;
		      } while (z.tam_int()>1);
else if (z.tam_int()<1) do
		      {
		      z.desl_esq(1);
		      z.e=z.e-1;
		      } while (z.tam_int()<1);
}
delete[] temp;
delete[] temp1;
return(z);
}

// Multiplies x by digit 'k'
Ereal mult_alg(Ereal x, char k)
{
Ereal z;

if (k==0) z.zera();
else
 {
 int i,t;
 char r,s;
 z.zera();
 if (k*x.d[0]<0) z.d[0]=-1;
 else if (k*x.d[0]>0) z.d[0]=1;
 else z.d[0]=0;
 i=max_int+x.tam_dec();
 r=0; s=0;
 t=x.tam_int();
 while (i>=max_int-t)
  {
  s=x.d[i]*k+r;
  z.d[i]=s % 10;
  r= s / 10;
  i--;
  }
 }
return(z);
}

// Note that this is a trial division. I implemented approximation
// methods but they proved no faster than this. You might try using
// such methods on your own. 
Ereal operator/(Ereal& a, Ereal& b)
{
int i,j;
int l;
char k;
Ereal x,y,r,s,t,u;
char chave;
int ponto;

if((a.e/2-b.e/2)>16382 || (a.e/2-b.e/2)<-16382) {
  throw TMathError("Number out of range.");
}
if (b.d[0]==0)
  {
  throw TMathError("Divide by zero.");
  }
r.zera();
s.zera();
t.zera();
if (a.d[0]==0) return(r);
r.d[0]=a.d[0]*b.d[0];
r.e=a.e-b.e;
x=a;
y=b;
x.d[0]=1;
y.d[0]=1;
x.e=0;
y.e=0;
chave=FALSE;
ponto=0;

if (x==y) {
	  r.d[max_int]=1;
	  ponto=1;
	  }
else if (x>y)
 {
 i=1;
 while (t<y)
  {
  t.desl_esq(1);
  t.d[max_tot]=x.d[max_int-x.tam_int()+i];
  t.d[0]=1;
  i++;
  }
 l=x.tam_int()-t.tam_int();
 i=1; j=1;
 chave=FALSE;
 do
 {
 k=10;
 do
  {
  k--;
  s.zera();
  s=mult_alg(y,k);
  } while (s>t);
 if ((ponto==0) && (j>l)) ponto=i;
 if ((k==0) && (chave==TRUE))
      {
      if (max_int-l+j<=max_tot) k=x.d[max_int-l+j];
			   else k=0;
      t.desce_alg(k);
      j++;
      r.d[i]=0;
      i++;
      chave=TRUE;
      }
 else if ((k==0) && (chave==FALSE))
      {
      chave=TRUE;
      if (max_int-l+j<=max_tot) k=x.d[max_int-l+j];
			   else k=0;
      t.desce_alg(k);
      j++;
      }
 else {
      r.d[i]=k;
      i++;
      u=subt_f(t,s);
      t=u;
      chave=FALSE;
      }
 } while (i<=max_tot);
 r.desl_dir(max_int-ponto);
 }
else if ((x<y) && (x.tam_int()+x.tam_dec()!=0))
 {
 i=0;
 t=x;
 while (t<y)
  {
  i++;
  t.desl_esq(1);
  }
 do
 {
 k=10;
 do
  {
  k--;
  s.zera();
  s=mult_alg(y,k);
  } while (s>t);
 if (k==0)
  {
  t.desce_alg(0);
  if (chave==TRUE) {
		   r.d[i]=0;
		   i++;
		   }
  chave=TRUE;
  }
 else
  {
  chave=FALSE;
  r.d[i]=k;
  i++;
  u=subt_f(t,s);
  t=u;
  }
 } while (i<=max_tot);
 r.desl_dir(max_int);
 }
if (r.d[0]!=0)
{
if (r.tam_int()>1) do
		      {
		      r.desl_dir(1);
		      r.e=r.e+1;
		      } while (r.tam_int()>1);
else if (r.tam_int()<1) do
		      {
		      r.desl_esq(1);
		      r.e=r.e-1;
		      } while (r.tam_int()<1);
}
return(r);
}

// Square root of x
Ereal raiz(Ereal& x)
{
int i,j;
char k;
Ereal r,s,t,u;

if (x.d[0]<0)
  {
  throw TMathError("Domain error.");
  }
r.zera();
if (x.d[0]!=0)
{
if (((x.e) % 2)!=0)
 {
 x.desl_esq(1);
 x.e=x.e-1;
 }
r.e=(x.e)/2;
s.zera();
t.zera();
u.zera();
i=max_int-1;
j=1;
do
 {
 if ((i+1)<max_tot) t.desce_2alg(x.d[i],x.d[i+1]);
	       else t.desl_esq(2);
 k=10;
 do
  {
  s=r;
  u=mult_alg(s,2);
  s=u;
  k--;
  s.desce_alg(k);
  u=mult_alg(s,k);
  s=u;
  u.zera();
  } while (s>t);
 r.desce_alg(k);
 u=subt_f(t,s);
 t=u;
 i=i+2;
 j++;
 } while (j<=max_dec);
r.desl_esq(1);
if (r.d[0]!=0)
{
if (r.tam_int()>1) do
		      {
		      r.desl_dir(1);
		      r.e=r.e+1;
		      } while (r.tam_int()>1);
else if (r.tam_int()<1) do
		      {
		      r.desl_esq(1);
		      r.e=r.e-1;
		      } while (r.tam_int()<1);
}
}
return(r);
}

// Absolute value of x (no signal)
Ereal absol(Ereal& x)
{
Ereal y;

y.zera();
y=x;
if (x.d[0]<0)
 {
 y.d[0]=-x.d[0];
 }
return(y);
}

// Integer part of x
Ereal inteiro(Ereal& x)
{
int i,k;
Ereal y;

y.zera();
k=x.e;
if (k>=0)
 {
 y=x;
 for (i=max_int+k+1;i<=max_tot;++i) y.d[i]=0;
 }
return(y);
}

// Fractional part of x
Ereal frac(Ereal& x)
{
int i,k;
Ereal y;

y.zera();
k=x.e;
y=x;
if (k>=0)
{
if ((k>=max_dec) || (k>=y.tam_dec())) y.zera();
else
 {
 for (i=1;i<=max_int+k;++i) y.d[i]=0;
 y.desl_esq(k+1);
 y.e=y.e-k-1;
 }
}
return(y);
}

// Raises x to the n'th power
Ereal eleva(Ereal& x, int n)
{
int i;
Ereal y,z;

z.zera();
z.d[0]=1;
z.d[max_int]=1;
if (n!=0)
 {
 for (i=1;i<=n;++i) {
		    y=z*x;
		    z=y;
		    }
 }
return(z);
}

// Converts x to double
double dbl(Ereal& x)
{
double y,z;
int i,e,sinal;

y=0; sinal=0; e=0;
if (x.d[0]!=0)
 {
 sinal=x.d[0];
 y=1; z=0;
 for (i=max_int;i>=1;--i)
  {
  z=z+x.d[i]*y;
  y=y*10;
  }
 y=10;
 for (i=1;i<=max_dec;++i)
  {
  z=z+x.d[max_int+i]/y;
  y=y*10;
  }
 e=x.e;
 if (e>0) {
  for (i=1;i<=e;++i) z=z*10;
  }
 else if (e<0) {
  e=-e;
  for (i=1;i<=e;++i) z=z/10;
  }
 }
 z=z*sinal;
 return(z);
}

// Converts x to Ereal
Ereal re(double x)
{
 Ereal y;
 char buf[TAM_DOUBLE];

 gcvt(x,DIGITOS_DOUBLE,buf);
 strupr(buf);
 y=r(buf);
 return(y);
}

// Converts x to int
int eint(Ereal& x)
{
 int i,z;
 static int dz[5] = { 1,10,100,1000,10000 };

 z=0;
 for (i=0;i<=x.e;++i)
   z+=x.d[max_int+i]*dz[x.e-i];
 z*=x.d[0];
 return(z);
}

// e^x (e=2.718281828...) 
Ereal exponencial(Ereal& x)
{
Ereal m,n,y,z,xn,w,eps0;

n.zera();
n=um;
z.zera();
y.zera();
y=um;
xn.zera();
xn=um;
w.zera();
m.zera();
 do
 {
 w=z+y;
 z=w;
 w=xn*x;
 xn=w;
 w=m+um;
 m=w;
 w=n*m;
 n=w;
 y=xn/n;
 eps0=absol(y);
 } while (eps0>eps);
return(z);
}

// Sine of x
Ereal seno(Ereal& x)
{
int i,snl;
Ereal m,n,y,z,xn,x2,w;

y.zera();
y=x;
snl=0;
if(y.e>max_dec)
  {
   throw TMathError("Total loss of significance.");
  }

if (y.d[0]<0) {
	       snl=-1;
	       y.d[0]=-y.d[0];
	       }
else if (y.d[0]>0) snl=1;

if (y>=pi_2)
 {
 Ereal v;

 v.zera();
 v=y/pi_2;
 v=inteiro(v);
 v=v*pi_2;
 y=y-v;
 }

if ((y>=pi1_2) && (y<pi))
  {
  y=pi-y;
  }
else if ((y>=pi) && (y<pi3_2))
  {
  y=y-pi;
  snl=-snl;
  }
else if ((y>=pi3_2) && (y<pi_2))
  {
  y=pi_2-y;
  snl=-snl;
  }

m.zera();
m=um;
n.zera();
n=um;
z.zera();
x2.zera();
x2=eleva(y,2);
xn.zera();
xn=y;
w.zera();

i=0;
 do
 {
 if ((i % 2) == 0) {
		   w=z+y;
		   z=w;
		   }
 else {
      w=z-y;
      z=w;
      }
 w=xn*x2;
 xn=w;
 i++;
 w=m+um;
 m=w;
 w=n*m;
 n=w;
 w=m+um;
 m=w;
 w=n*m;
 n=w;
 y=xn/n;
 } while (y>eps);
z.d[0]=snl*z.d[0];
return(z);
}

// Co-sine of x
Ereal cosseno(Ereal& x)
{
int i,snl;
Ereal m,n,y,z,xn,x2,w;

y.zera();
y=x;
snl=1;
if(y.e>max_dec)
  {
   throw TMathError("Total loss of significance.");
  }
if (y.d[0]<0) {
	       y.d[0]=-y.d[0];
	       }

if (y>=pi_2)
 {
 Ereal v;

 v.zera();
 v=y/pi_2;
 v=inteiro(v);
 v=v*pi_2;
 y=y-v;
 }

if ((y>=pi1_2) && (y<pi))
  {
  y=pi-y;
  snl=-snl;
  }
else if ((y>=pi) && (y<pi3_2))
  {
  y=y-pi;
  snl=-snl;
  }
else if ((y>=pi3_2) && (y<pi_2))
  {
  y=pi_2-y;
  }

n.zera();
n=um;
z.zera();
x2.zera();
x2=eleva(y,2);
y.zera();
y=um;
xn.zera();
xn=um;
w.zera();
m.zera();
i=0;
 do
 {
 if ((i % 2) == 0) {
		   w=z+y;
		   z=w;
		   }
 else {
      w=z-y;
      z=w;
      }
 w=xn*x2;
 xn=w;
 i++;
 w=m+um;
 m=w;
 w=n*m;
 n=w;
 w=m+um;
 m=w;
 w=n*m;
 n=w;
 y=xn/n;
 } while (y>eps);
z.d[0]=snl*z.d[0];
return(z);
}

// Inverse tangent of x
Ereal arctg(Ereal& x)
{
Ereal n,y,z,xn,x2,dois("2"),r_menosum("-1"),eps0;

if(x>=r_menosum && x<=um) {
  xn=x;
  x2=-eleva(x,2);
  n=um;
  z.zera();
  do {
    y = xn/n;
    z += y;
    xn *= x2;
    n += dois;
    eps0=absol(y);
  } while(eps0>eps);
}
else {
  xn=um/x;
  x2=-eleva(x,2);
  n=um;
  z.zera();
  do {
    y = xn/n;
    z += y;
    xn *= x2;
    n += dois;
    eps0=absol(y);
  } while(eps0>eps);
  z = pi1_2-z;
}
return(z);
}

// Factorial of n
Ereal fatorial(int& n)
{
int i;
Ereal k,t,x;
x.zera();
k.zera();
t.zera();
x=r("1");
if(n<0) {
  throw TMathError("! of negative number.");
}
else if (n!=0)
 {
 for (i=1;i<=n;++i)
  {
  t=k+um;
  k=t;
  t=x*k;
  x=t;
  }
 }
return x;
}
